#include <iostream>

#include <TStopwatch.h>
#include <TProfile.h>
#include <TH1F.h>
#include <TF1.h>
#include <TFile.h>
#include <TMath.h>

#include "mfNamespace.h"

using mfEp::NetaGaps;
using mfRes::NresBin;
using mfRes::NresChiIter;
using mfRes::ResChi;
using mfRes::ResEventPlane;

using TMath::Sqrt;

int main(int argc, char **argv)
{
  TString iFileName, oFileName;

  if (argc < 5)
  {
    std::cerr << "./fit-res -i INPUT_RES_HIST.root -o OUTPUT_RES_FIT.root" << std::endl;
    return 1;
  }
  for (int i = 1; i < argc; i++)
  {
    if (std::string(argv[i]) != "-i" &&
        std::string(argv[i]) != "-o")
    {
      std::cerr << "\n[ERROR]: Unknown parameter " << i << ": " << argv[i] << std::endl;
      return 2;
    }
    else
    {
      if (std::string(argv[i]) == "-i" && i != argc - 1)
      {
        iFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-i" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-o" && i != argc - 1)
      {
        oFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-o" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Output file name was not specified " << std::endl;
        return 1;
      }
    }
  }

  TStopwatch timer;
  timer.Start();

  // Init input profiles
  TProfile *pResRecoSqrTpc[NetaGaps];
  TProfile *pResRecoSqrFHCal;

  TFile *fi = new TFile(iFileName.Data(), "READ");
  fi->cd();

  for (int i = 0; i < NetaGaps; i++)
  {
    pResRecoSqrTpc[i] = (TProfile *)fi->Get(Form("pResRecoSqrtTpc_gap%i", i));
  }
  pResRecoSqrFHCal = (TProfile *)fi->Get(Form("pResRecoSqrFHCal"));

  // Init output histograms
  TH1F *hResTpc[NetaGaps];
  TH1F *hResV1FHCal;
  TH1F *hResV2FHCal;

  TF1 *fResTpc[NetaGaps];
  TF1 *fResV1FHCal;
  TF1 *fResV2FHCal;

  for (int i = 0; i < NetaGaps; i++)
  {
    hResTpc[i] = new TH1F(Form("hResTpc_gap%i", i),
                          Form("hResTpc_gap%i", i),
                          NresBin, 0., 100.);
    fResTpc[i] = new TF1(Form("fResTpc_gap%i", i),
                         "pol6",
                         0.01, 80.);
  }
  hResV1FHCal = new TH1F(Form("hResV1FHCal"),
                         Form("hResV1FHCal"),
                         NresBin, 0., 100.);
  hResV2FHCal = new TH1F(Form("hResV2FHCal"),
                         Form("hResV2FHCal"),
                         NresBin, 0., 100.);
  fResV1FHCal = new TF1(Form("fResV1FHCal"),
                        "pol8",
                        0.01, 80.);
  fResV2FHCal = new TF1(Form("fResV2FHCal"),
                        "pol8",
                        0.01, 80.);

  // Calculate resolution for Tpc EP
  for (int iGap = 0; iGap < NetaGaps; iGap++)
  {
    for (int iBin = 0; iBin < pResRecoSqrTpc[iGap]->GetNbinsX(); iBin++)
    {
      double res2 = pResRecoSqrTpc[iGap]->GetBinContent(iBin + 1);
      double Eres2 = pResRecoSqrTpc[iGap]->GetBinError(iBin + 1);
      if (res2 <= 0)
        continue;
      double res = Sqrt(res2);
      double Eres = Eres2 / (2 * res2);
      hResTpc[iGap]->SetBinContent(iBin + 1, res);
      hResTpc[iGap]->SetBinError(iBin + 1, Eres);
    }
  }

  // Calculate resolution for FHCal EP
  for (int iBin = 0; iBin < pResRecoSqrFHCal->GetNbinsX(); iBin++)
  {
    double res2 = pResRecoSqrFHCal->GetBinContent(iBin + 1);
    if (res2 <= 0)
      continue;
    double res = Sqrt(res2);

    double chi_half_v1 = ResChi(res, 1., NresChiIter);
    double chi_v1 = Sqrt(2) * chi_half_v1;

    double res_v1 = ResEventPlane(chi_v1, 1.);
    double res_v2 = ResEventPlane(chi_v1, 2.);

    hResV1FHCal->SetBinContent(iBin + 1, res_v1);
    hResV1FHCal->SetBinError(iBin + 1, 0.);
    hResV2FHCal->SetBinContent(iBin + 1, res_v2);
    hResV2FHCal->SetBinError(iBin + 1, 0.);
  }

  // Fit resolutions
  for (int iGap = 0; iGap < NetaGaps; iGap++)
  {
    hResTpc[iGap]->Fit(fResTpc[iGap],"WR");
  }
  hResV1FHCal->Fit(fResV1FHCal,"WR");
  hResV2FHCal->Fit(fResV2FHCal,"WR");

  // Write output
  TFile *fo = new TFile(oFileName.Data(), "recreate");
  fo->cd();

  for (int i = 0; i < NetaGaps; i++)
  {
    hResTpc[i]->Write();
  }
  hResV1FHCal->Write();
  hResV2FHCal->Write();
  for (int i = 0; i < NetaGaps; i++)
  {
    fResTpc[i]->Write();
  }
  fResV1FHCal->Write();
  fResV2FHCal->Write();

  fo->Close();

  timer.Stop();
  timer.Print();
  return 0;
}