#include "mfPicoDstReader.h"
#include "mfNamespace.h"

#include <TStopwatch.h>
#include <TH1F.h>
#include <TClonesArray.h>
#include <TFile.h>
#include <iostream>

#include <PicoDstMCEvent.h>
#include <PicoDstRecoEvent.h>
#include <PicoDstMCTrack.h>
#include <PicoDstRecoTrack.h>
#include <PicoDstFHCal.h>

using mfdca::etaBins;
using mfdca::Ndim;
using mfdca::NetaBins;
using mfdca::NptBins;
using mfdca::ptBins;

using mfdca::GetEtaBin;
using mfdca::GetPtBin;

int main(int argc, char **argv)
{
  TString iFileName, oFileName;

  if (argc < 5)
  {
    std::cerr << "./get-dca -i INPUTFILE/LIST -o OUTPUT.root" << std::endl;
    return 1;
  }
  for (int i = 1; i < argc; i++)
  {
    if (std::string(argv[i]) != "-i" &&
        std::string(argv[i]) != "-o")
    {
      std::cerr << "\n[ERROR]: Unknown parameter " << i << ": " << argv[i] << std::endl;
      return 2;
    }
    else
    {
      if (std::string(argv[i]) == "-i" && i != argc - 1)
      {
        iFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-i" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-o" && i != argc - 1)
      {
        oFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-o" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Output file name was not specified " << std::endl;
        return 1;
      }
    }
  }

  TStopwatch timer;
  timer.Start();

  mfPicoDstReader *input = new mfPicoDstReader();
  TChain *inChain = (TChain *)input->Open(iFileName);

  if (!inChain)
  {
    std::cerr << "[ERROR]: nothing to read. Maybe the input TTree(s) is(are) empty?" << std::endl;
    return 10;
  }

  // Init output histograms
  TH1F *h_dca[Ndim][NptBins][NetaBins];

  for (int dim = 0; dim < Ndim; ++dim)
  {
    for (int ptbin = 0; ptbin < NptBins; ++ptbin)
    {
      for (int etabin = 0; etabin < NetaBins; ++etabin)
      {
        h_dca[dim][ptbin][etabin] = new TH1F(
            Form("h_dca_%i_%i_%i", dim, ptbin, etabin),
            Form("DCA distribution for %.2f < p_{t} < %.2f and %.2f < #eta < %.2f",
                 ptBins[ptbin], ptBins[ptbin + 1], etaBins[etabin], etaBins[etabin + 1]),
            4000, -50., 50);
      }
    }
  }

  // PicoDstMCEvent *mcEvent = nullptr;
  // PicoDstRecoEvent *recoEvent = nullptr;
  TClonesArray *recoTracks = nullptr;
  // TClonesArray *mcTracks = nullptr;
  // TClonesArray *fhcalmodules = nullptr;

  // inChain->SetBranchAddress("mcevent.", &mcEvent);
  // inChain->SetBranchAddress("recoevent.", &recoEvent);
  // inChain->SetBranchAddress("mctracks", &mcTracks);
  inChain->SetBranchAddress("recotracks", &recoTracks);
  // inChain->SetBranchAddress("FHCalModules", &fhcalmodules);

  // Start event loop
  int n_entries = inChain->GetEntries();
  for (int iEv = 0; iEv < n_entries; iEv++)
  {
    if (iEv % 1000 == 0)
      std::cout << "Event [" << iEv << "/" << n_entries << "]" << std::endl;
    inChain->GetEntry(iEv);

    Int_t reco_mult = recoTracks->GetEntriesFast();
    // Read Reco tracks
    for (int iTr = 0; iTr < reco_mult; iTr++)
    {
      auto recoTrack = (PicoDstRecoTrack *)recoTracks->UncheckedAt(iTr);
      Int_t pt_bin = GetPtBin(TMath::Abs(recoTrack->GetPt()));
      Int_t eta_bin = GetEtaBin(recoTrack->GetEta());
      if ((eta_bin == -1) || (pt_bin == -1))
        continue;
      h_dca[0][pt_bin][eta_bin]->Fill(recoTrack->GetDCAx());
      h_dca[1][pt_bin][eta_bin]->Fill(recoTrack->GetDCAy());
      h_dca[2][pt_bin][eta_bin]->Fill(recoTrack->GetDCAz());
    }
  }

  // Init output file
  TFile *fo = new TFile(oFileName.Data(), "recreate");
  fo->cd();
  for (int dim = 0; dim < Ndim; ++dim)
  {
    for (int ptbin = 0; ptbin < NptBins; ++ptbin)
    {
      for (int etabin = 0; etabin < NetaBins; ++etabin)
      {
        // std::cout << h_dca[dim][ptbin][etabin]->GetName() << " ("
        //           << h_dca[dim][ptbin][etabin]->GetEntries() << " entries)"
        //           << std::endl;
        h_dca[dim][ptbin][etabin]->Write();
      }
    }
  }
  std::cout << "Closing output file." << std::endl;
  fo->Close();

  timer.Stop();
  timer.Print();
  return 0;
}