#include "mfPicoDstReader.h"
#include "mfNamespace.h"
#include "mfProcessEvent.h"
#include "mfInputReader.h"
#include "mfSelector.h"

#include <TStopwatch.h>
#include <TProfile2D.h>
#include <TClonesArray.h>
#include <TFile.h>
#include <iostream>
#include <TMath.h>

#include <PicoDstMCEvent.h>
#include <PicoDstRecoEvent.h>
#include <PicoDstMCTrack.h>
#include <PicoDstRecoTrack.h>
#include <PicoDstFHCal.h>

using mfdca::Ndim;
using mfEp::EpNames;
using mfEp::NetaGaps;
using mfEp::TpcEtaGap;
using mfEp::TpcLEp;
using mfEp::TpcREp;
using mfEp::EpNames::kFHCal;
using mfEp::EpNames::kFHCal_L;
using mfEp::EpNames::kFHCal_R;
using mfFlow::MinV1PtEta;
using mfPid::Mass2;
using mfPid::MinPidProbability;
using mfPid::Npid;
using mfRes::NresBin;

using TMath::Abs;
using TMath::ATan2;
using TMath::Cos;
using TMath::Log;
using TMath::Pi;
using TMath::Power;
using TMath::Sin;
using TMath::Sqrt;

int main(int argc, char **argv)
{
  TString iFileName, oFileName, dcaFileName, resFileName;

  if (argc < 9)
  {
    std::cerr << "./get-flow -i INPUTFILE/LIST -o OUTPUT.root -dca DCA_FIT.root -res RES_FIT.root" << std::endl;
    return 1;
  }
  for (int i = 1; i < argc; i++)
  {
    if (std::string(argv[i]) != "-i" &&
        std::string(argv[i]) != "-o" &&
        std::string(argv[i]) != "-dca" &&
        std::string(argv[i]) != "-res")
    {
      std::cerr << "\n[ERROR]: Unknown parameter " << i << ": " << argv[i] << std::endl;
      return 2;
    }
    else
    {
      if (std::string(argv[i]) == "-i" && i != argc - 1)
      {
        iFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-i" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-o" && i != argc - 1)
      {
        oFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-o" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Output file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-dca" && i != argc - 1)
      {
        dcaFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-dca" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-res" && i != argc - 1)
      {
        resFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-res" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
    }
  }

  TStopwatch timer;
  timer.Start();

  mfPicoDstReader *input = new mfPicoDstReader();
  TChain *inChain = (TChain *)input->Open(iFileName);

  if (!inChain)
  {
    std::cerr << "[ERROR]: nothing to read. Maybe the input TTree(s) is(are) empty?" << std::endl;
    return 10;
  }

  // Init output profiles
  TProfile2D *pv2PtTpc[NetaGaps][Npid];
  TProfile2D *pv1PtFHCal[Npid];
  TProfile2D *pv2PtFHCal[Npid];
  TProfile2D *pv2YTpc[NetaGaps][Npid];
  TProfile2D *pv1YFHCal[Npid];
  TProfile2D *pv2YFHCal[Npid];

  for (int iGap = 0; iGap < NetaGaps; iGap++)
  {
    for (int iPid = 0; iPid < Npid; iPid++)
    {
      pv2PtTpc[iGap][iPid] = new TProfile2D(Form("pv2PtTpc_gap%i_pid%i", iGap, iPid),
                                            Form("pv2PtTpc_gap%i_pid%i", iGap, iPid),
                                            300, 0., 3., NresBin, 0., 100.);
      pv2YTpc[iGap][iPid] = new TProfile2D(Form("pv2YTpc_gap%i_pid%i", iGap, iPid),
                                           Form("pv2YTpc_gap%i_pid%i", iGap, iPid),
                                           320, -1.6, 1.6, NresBin, 0., 100.);
    }
  }
  for (int iPid = 0; iPid < Npid; iPid++)
  {
    pv1PtFHCal[iPid] = new TProfile2D(Form("pv1PtFHCal_pid%i", iPid),
                                      Form("pv1PtFHCal_pid%i", iPid),
                                      300, 0., 3., NresBin, 0., 100.);
    pv2PtFHCal[iPid] = new TProfile2D(Form("pv2PtFHCal_pid%i", iPid),
                                      Form("pv2PtFHCal_pid%i", iPid),
                                      300, 0., 3., NresBin, 0., 100.);
    pv1YFHCal[iPid] = new TProfile2D(Form("pv1YFHCal_pid%i", iPid),
                                     Form("pv1YFHCal_pid%i", iPid),
                                     320, -1.6, 1.6, NresBin, 0., 100.);
    pv2YFHCal[iPid] = new TProfile2D(Form("pv2YFHCal_pid%i", iPid),
                                     Form("pv2YFHCal_pid%i", iPid),
                                     320, -1.6, 1.6, NresBin, 0., 100.);
  }

  PicoDstMCEvent *mcEvent = nullptr;
  // PicoDstRecoEvent *recoEvent = nullptr;
  TClonesArray *recoTracks = nullptr;
  // TClonesArray *mcTracks = nullptr;
  TClonesArray *fhcalmodules = nullptr;

  inChain->SetBranchAddress("mcevent.", &mcEvent);
  // inChain->SetBranchAddress("recoevent.", &recoEvent);
  // inChain->SetBranchAddress("mctracks", &mcTracks);
  inChain->SetBranchAddress("recotracks", &recoTracks);
  inChain->SetBranchAddress("FHCalModules", &fhcalmodules);

  TF1 *fResTpc[NetaGaps];
  TF1 *fResv1FHCal, *fResv2FHCal;

  TFile *fiRes = new TFile(resFileName.Data(), "READ");
  fiRes->cd();

  for (int i = 0; i < NetaGaps; i++)
  {
    fResTpc[i] = (TF1 *)fiRes->Get(Form("fResTpc_gap%i", i));
  }
  fResv1FHCal = (TF1 *)fiRes->Get(Form("fResV1FHCal"));
  fResv2FHCal = (TF1 *)fiRes->Get(Form("fResV2FHCal"));

  std::vector<float> vPsiTpcR, vPsiTpcL;
  float PsiFHCal;

  // Start event loop
  int n_entries = inChain->GetEntries();
  mfSelector *selector = new mfSelector();
  mfInputReader *inputReader = new mfInputReader();
  inputReader->SetDcaFitFile(dcaFileName);
  for (int iEv = 0; iEv < n_entries; iEv++)
  {
    if (iEv % 1000 == 0)
      std::cout << "Event [" << iEv << "/" << n_entries << "]" << std::endl;
    inChain->GetEntry(iEv);

    mfProcessEvent *procEvent = new mfProcessEvent();
    float cent = procEvent->GetCentB(mcEvent);
    procEvent->SetInputReader(inputReader);

    // std::vector<double> vResTpc;
    // for (int iGap=0; iGap<NetaGaps; iGap++)
    // {
    //   vResTpc.push_back(fResTpc[iGap]->Eval(cent));
    // }

    // std::cout << "\tNtracks = " << recoTracks->GetEntriesFast() << std::endl;

    mfEventInfo evInfo = procEvent->ProcessRecoEP(recoTracks, fhcalmodules);
    vPsiTpcR.clear();
    vPsiTpcL.clear();
    PsiFHCal = 0.;
    if (evInfo.GetQvectors().size() == 0)
      continue;

    for (int iGap = 0; iGap < NetaGaps; iGap++)
    {
      // Ntracks >= 4 for TpcR and TpcL EP
      if (selector->isGoodTpcEp(evInfo.GetEpMult(TpcREp[iGap]), evInfo.GetEpMult(TpcLEp[iGap])))
      {
        vPsiTpcR.push_back(evInfo.GetQvector(TpcREp[iGap]).GetPhi());
        vPsiTpcL.push_back(evInfo.GetQvector(TpcLEp[iGap]).GetPhi());
      }
      else
      {
        vPsiTpcR.push_back(-999.);
        vPsiTpcL.push_back(-999.);
      }
    }

    // FHCal energy > 0 in FHCal_L and FHCal_R
    if (selector->isGoodFHCalEp(evInfo.GetEpWeight(kFHCal_R), evInfo.GetEpWeight(kFHCal_R)))
      PsiFHCal = evInfo.GetQvector(kFHCal).GetPhi();
    else
      PsiFHCal = -999.;

    // Starting flow calculation
    for (int iTrack = 0; iTrack < recoTracks->GetEntriesFast(); iTrack++)
    {
      auto recoTrack = (PicoDstRecoTrack *)recoTracks->UncheckedAt(iTrack);
      if (!recoTrack)
        continue;

      // Primary track selection
      if (!selector->isPrimary(recoTrack->GetDCAx(), inputReader->GetDcaSigmaX(recoTrack)))
        continue;
      if (!selector->isPrimary(recoTrack->GetDCAy(), inputReader->GetDcaSigmaY(recoTrack)))
        continue;
      if (!selector->isPrimary(recoTrack->GetDCAz(), inputReader->GetDcaSigmaZ(recoTrack)))
        continue;

      // Good reco track selection
      if (!selector->isGoodRecoTrack(recoTrack))
        continue;

      double Rapidity, Energy = -1;
      if (selector->isPion(recoTrack))
        Energy = Sqrt(Power(recoTrack->GetP(), 2) + Mass2[0]);
      if (selector->isKaon(recoTrack))
        Energy = Sqrt(Power(recoTrack->GetP(), 2) + Mass2[1]);
      if (selector->isProton(recoTrack))
        Energy = Sqrt(Power(recoTrack->GetP(), 2) + Mass2[2]);

      if (Energy != -1)
        Rapidity = 0.5 * Log((Energy + recoTrack->GetPz()) / (Energy - recoTrack->GetPz()));
      else
        Rapidity = -999.;

      // v2 (TPC EP)
      for (int iGap = 0; iGap < NetaGaps; iGap++)
      {
        double v2Tpc;
        if (vPsiTpcR.at(iGap) == -999.)
          continue;
        if (recoTrack->GetEta() < -1. * TpcEtaGap[iGap])
        {
          v2Tpc = Cos(2. * (recoTrack->GetPhi() - vPsiTpcR.at(iGap))) / fResTpc[iGap]->Eval(cent);

          pv2PtTpc[iGap][0]->Fill(recoTrack->GetPt(), cent, v2Tpc);
          if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
            pv2PtTpc[iGap][1]->Fill(recoTrack->GetPt(), cent, v2Tpc);
          if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
            pv2PtTpc[iGap][2]->Fill(recoTrack->GetPt(), cent, v2Tpc);
          if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
            pv2PtTpc[iGap][3]->Fill(recoTrack->GetPt(), cent, v2Tpc);

          if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
            pv2YTpc[iGap][1]->Fill(Rapidity, cent, v2Tpc);
          if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
            pv2YTpc[iGap][2]->Fill(Rapidity, cent, v2Tpc);
          if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
            pv2YTpc[iGap][3]->Fill(Rapidity, cent, v2Tpc);
        }
        if (recoTrack->GetEta() > TpcEtaGap[iGap])
        {
          v2Tpc = Cos(2. * (recoTrack->GetPhi() - vPsiTpcL.at(iGap))) / fResTpc[iGap]->Eval(cent);

          pv2PtTpc[iGap][0]->Fill(recoTrack->GetPt(), cent, v2Tpc);
          if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
            pv2PtTpc[iGap][1]->Fill(recoTrack->GetPt(), cent, v2Tpc);
          if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
            pv2PtTpc[iGap][2]->Fill(recoTrack->GetPt(), cent, v2Tpc);
          if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
            pv2PtTpc[iGap][3]->Fill(recoTrack->GetPt(), cent, v2Tpc);

          if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
            pv2YTpc[iGap][1]->Fill(Rapidity, cent, v2Tpc);
          if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
            pv2YTpc[iGap][2]->Fill(Rapidity, cent, v2Tpc);
          if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
            pv2YTpc[iGap][3]->Fill(Rapidity, cent, v2Tpc);
        }
      }

      // v2 FHCal
      double v1FHCal, v2FHCal, w_sign;
      if (PsiFHCal != -999.)
      {
        w_sign = (recoTrack->GetEta() > 0) ? 1 : -1;
        v1FHCal = w_sign * Cos(1. * (recoTrack->GetPhi() - PsiFHCal)) / fResv1FHCal->Eval(cent);
        v2FHCal = Cos(2. * (recoTrack->GetPhi() - PsiFHCal)) / fResv2FHCal->Eval(cent);

        if (Abs(recoTrack->GetEta()) > MinV1PtEta)
        {
          pv1PtFHCal[0]->Fill(recoTrack->GetPt(), cent, v1FHCal);
          if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
            pv1PtFHCal[1]->Fill(recoTrack->GetPt(), cent, v1FHCal);
          if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
            pv1PtFHCal[2]->Fill(recoTrack->GetPt(), cent, v1FHCal);
          if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
            pv1PtFHCal[3]->Fill(recoTrack->GetPt(), cent, v1FHCal);
        }
        
        if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
          pv1YFHCal[1]->Fill(Rapidity, cent, v1FHCal);
        if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
          pv1YFHCal[2]->Fill(Rapidity, cent, v1FHCal);
        if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
          pv1YFHCal[3]->Fill(Rapidity, cent, v1FHCal);

        pv2PtFHCal[0]->Fill(recoTrack->GetPt(), cent, v2FHCal);
        if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
          pv2PtFHCal[1]->Fill(recoTrack->GetPt(), cent, v2FHCal);
        if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
          pv2PtFHCal[2]->Fill(recoTrack->GetPt(), cent, v2FHCal);
        if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
          pv2PtFHCal[3]->Fill(recoTrack->GetPt(), cent, v2FHCal);

        if (selector->isPion(recoTrack) && recoTrack->GetCharge() > 0)
          pv2YFHCal[1]->Fill(Rapidity, cent, v2FHCal);
        if (selector->isKaon(recoTrack) && recoTrack->GetCharge() > 0)
          pv2YFHCal[2]->Fill(Rapidity, cent, v2FHCal);
        if (selector->isProton(recoTrack) && recoTrack->GetCharge() > 0)
          pv2YFHCal[3]->Fill(Rapidity, cent, v2FHCal);
      }
    }
  }

  // Init output file
  TFile *fo = new TFile(oFileName.Data(), "recreate");
  fo->cd();
  for (int iGap = 0; iGap < NetaGaps; iGap++)
  {
    for (int iPid = 0; iPid < Npid; iPid++)
    {
      pv2PtTpc[iGap][iPid]->Write();
    }
  }
  for (int iPid = 0; iPid < Npid; iPid++)
  {
    pv1PtFHCal[iPid]->Write();
  }
  for (int iPid = 0; iPid < Npid; iPid++)
  {
    pv2PtFHCal[iPid]->Write();
  }

  for (int iGap = 0; iGap < NetaGaps; iGap++)
  {
    for (int iPid = 0; iPid < Npid; iPid++)
    {
      pv2YTpc[iGap][iPid]->Write();
    }
  }
  for (int iPid = 0; iPid < Npid; iPid++)
  {
    pv1YFHCal[iPid]->Write();
  }
  for (int iPid = 0; iPid < Npid; iPid++)
  {
    pv2YFHCal[iPid]->Write();
  }
  fo->Close();

  timer.Stop();
  timer.Print();
  return 0;
}