#include "mfPicoDstReader.h"
#include "mfNamespace.h"
#include "mfProcessEvent.h"
#include "mfInputReader.h"

#include <TStopwatch.h>
#include <TH2F.h>
#include <TClonesArray.h>
#include <TFile.h>
#include <iostream>
#include <TMath.h>

#include <PicoDstMCEvent.h>
#include <PicoDstRecoEvent.h>
#include <PicoDstMCTrack.h>
#include <PicoDstRecoTrack.h>
#include <PicoDstFHCal.h>

using mfdca::Ndim;
using mfEp::EpNames;
using mfEp::NetaGaps;
using mfEp::NsubEvents;
using mfEp::RangeAngle;
using mfEp::TpcEtaGap;
using mfEp::TpcLEp;
using mfEp::TpcREp;
using mfEp::EpNames::kFHCal;
using mfEp::EpNames::kFHCal_L;
using mfEp::EpNames::kFHCal_R;
using mfRes::NresBin;

using TMath::Cos;
using TMath::Pi;
using TMath::Sin;

int main(int argc, char **argv)
{
  TString iFileName, oFileName, dcaFileName;

  if (argc < 7)
  {
    std::cerr << "./get-qv -i INPUTFILE/LIST -o OUTPUT.root -dca DCA_FIT.root" << std::endl;
    return 1;
  }
  for (int i = 1; i < argc; i++)
  {
    if (std::string(argv[i]) != "-i" &&
        std::string(argv[i]) != "-o" &&
        std::string(argv[i]) != "-dca")
    {
      std::cerr << "\n[ERROR]: Unknown parameter " << i << ": " << argv[i] << std::endl;
      return 2;
    }
    else
    {
      if (std::string(argv[i]) == "-i" && i != argc - 1)
      {
        iFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-i" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-o" && i != argc - 1)
      {
        oFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-o" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Output file name was not specified " << std::endl;
        return 1;
      }
      if (std::string(argv[i]) == "-dca" && i != argc - 1)
      {
        dcaFileName = argv[++i];
        continue;
      }
      if (std::string(argv[i]) == "-dca" && i == argc - 1)
      {
        std::cerr << "\n[ERROR]: Input file name was not specified " << std::endl;
        return 1;
      }
    }
  }

  TStopwatch timer;
  timer.Start();

  mfPicoDstReader *input = new mfPicoDstReader();
  TChain *inChain = (TChain *)input->Open(iFileName);

  if (!inChain)
  {
    std::cerr << "[ERROR]: nothing to read. Maybe the input TTree(s) is(are) empty?" << std::endl;
    return 10;
  }

  // Init output profiles
  TH2F *hQxVsCent[NsubEvents];
  TH2F *hQyVsCent[NsubEvents];
  TH2F *hPsiEpVsCent[NsubEvents];
  TH2F *hQxRLVsCent[NetaGaps + 1];
  TH2F *hQyRLVsCent[NetaGaps + 1];
  TH2F *hPsiEpRLVsCent[NetaGaps + 1];
  TH2F *hPsiRpRVsCent[NetaGaps + 1];
  TH2F *hPsiRpLVsCent[NetaGaps + 1];
  for (int i = 0; i < NsubEvents; i++)
  {
    hQxVsCent[i] = new TH2F(Form("hQxVsCent_%i", i),
                            Form("hQxVsCent_%i", i),
                            600, -1.1, 1.1, NresBin, 0., 100.);
    hQyVsCent[i] = new TH2F(Form("hQyVsCent_%i", i),
                            Form("hQyVsCent_%i", i),
                            600, -1.1, 1.1, NresBin, 0., 100.);
    hPsiEpVsCent[i] = new TH2F(Form("hPsiEpVsCent_%i", i),
                               Form("hPsiEpVsCent_%i", i),
                               360, 0., 2 * Pi(), NresBin, 0., 100.);
  }
  for (int i = 0; i < NetaGaps + 1; i++)
  {
    hQxRLVsCent[i] = new TH2F(Form("hQxRLVsCent_%i", i),
                              Form("hQxRLVsCent_%i", i),
                              600, -1.1, 1.1, NresBin, 0., 100.);
    hQyRLVsCent[i] = new TH2F(Form("hQyRLVsCent_%i", i),
                              Form("hQyRLVsCent_%i", i),
                              600, -1.1, 1.1, NresBin, 0., 100.);
    hPsiEpRLVsCent[i] = new TH2F(Form("hPsiEpRLVsCent_%i", i),
                                 Form("hPsiEpRLVsCent_%i", i),
                                 360, -Pi(), Pi(), NresBin, 0., 100.);
    hPsiRpRVsCent[i] = new TH2F(Form("hPsiRpRVsCent_%i", i),
                                Form("hPsiRpRVsCent_%i", i),
                                360, -Pi(), Pi(), NresBin, 0., 100.);
    hPsiRpLVsCent[i] = new TH2F(Form("hPsiRpLVsCent_%i", i),
                                Form("hPsiRpLVsCent_%i", i),
                                360, -Pi(), Pi(), NresBin, 0., 100.);
  }

  PicoDstMCEvent *mcEvent = nullptr;
  // PicoDstRecoEvent *recoEvent = nullptr;
  TClonesArray *recoTracks = nullptr;
  // TClonesArray *mcTracks = nullptr;
  TClonesArray *fhcalmodules = nullptr;

  inChain->SetBranchAddress("mcevent.", &mcEvent);
  // inChain->SetBranchAddress("recoevent.", &recoEvent);
  // inChain->SetBranchAddress("mctracks", &mcTracks);
  inChain->SetBranchAddress("recotracks", &recoTracks);
  inChain->SetBranchAddress("FHCalModules", &fhcalmodules);

  // Start event loop
  int n_entries = inChain->GetEntries();
  mfInputReader *inputReader = new mfInputReader();
  inputReader->SetDcaFitFile(dcaFileName);
  for (int iEv = 0; iEv < n_entries; iEv++)
  {
    if (iEv % 1000 == 0)
      std::cout << "Event [" << iEv << "/" << n_entries << "]" << std::endl;
    inChain->GetEntry(iEv);

    mfProcessEvent *procEvent = new mfProcessEvent();
    float cent = procEvent->GetCentB(mcEvent);
    procEvent->SetInputReader(inputReader);

    // std::cout << "\tNtracks = " << recoTracks->GetEntriesFast() << std::endl;

    mfEventInfo evInfo = procEvent->ProcessRecoEP(recoTracks, fhcalmodules);
    if (evInfo.GetQvectors().size() == 0)
      continue;

    for (int iGap = 0; iGap < NetaGaps; iGap++)
    {
      // Ntracks >= 2 for TpcR EP
      if (evInfo.GetEpMult(TpcREp[iGap]) < 2)
        continue;
      // Ntracks >= 2 for TpcL EP
      if (evInfo.GetEpMult(TpcLEp[iGap]) < 2)
        continue;

      hQxVsCent[TpcREp[iGap]]->Fill(evInfo.GetQvector(TpcREp[iGap]).X(), cent);
      hQyVsCent[TpcREp[iGap]]->Fill(evInfo.GetQvector(TpcREp[iGap]).Y(), cent);
      hQxVsCent[TpcLEp[iGap]]->Fill(evInfo.GetQvector(TpcLEp[iGap]).X(), cent);
      hQyVsCent[TpcLEp[iGap]]->Fill(evInfo.GetQvector(TpcLEp[iGap]).Y(), cent);

      hPsiEpVsCent[TpcREp[iGap]]->Fill(evInfo.GetQvector(TpcREp[iGap]).GetPhi(), cent);
      hPsiEpVsCent[TpcLEp[iGap]]->Fill(evInfo.GetQvector(TpcLEp[iGap]).GetPhi(), cent);

      hQxRLVsCent[iGap]->Fill(evInfo.GetQvector(TpcREp[iGap]).X() - evInfo.GetQvector(TpcLEp[iGap]).X(), cent);
      hQyRLVsCent[iGap]->Fill(evInfo.GetQvector(TpcREp[iGap]).Y() - evInfo.GetQvector(TpcLEp[iGap]).Y(), cent);

      hPsiEpRLVsCent[iGap]->Fill(RangeAngle(evInfo.GetQvector(TpcREp[iGap]).GetPhi() - evInfo.GetQvector(TpcLEp[iGap]).GetPhi()), cent);
      hPsiRpRVsCent[iGap]->Fill(RangeAngle(evInfo.GetQvector(TpcREp[iGap]).GetPhi() - mcEvent->GetPhiRP()), cent);
      hPsiRpLVsCent[iGap]->Fill(RangeAngle(evInfo.GetQvector(TpcLEp[iGap]).GetPhi() - mcEvent->GetPhiRP()), cent);
    }

    hQxVsCent[kFHCal_R]->Fill(evInfo.GetQvector(kFHCal_R).X(), cent);
    hQyVsCent[kFHCal_R]->Fill(evInfo.GetQvector(kFHCal_R).Y(), cent);
    hQxVsCent[kFHCal_L]->Fill(evInfo.GetQvector(kFHCal_L).X(), cent);
    hQyVsCent[kFHCal_L]->Fill(evInfo.GetQvector(kFHCal_L).Y(), cent);
    hQxVsCent[kFHCal]->Fill(evInfo.GetQvector(kFHCal).X(), cent);
    hQyVsCent[kFHCal]->Fill(evInfo.GetQvector(kFHCal).Y(), cent);

    hPsiEpVsCent[kFHCal_R]->Fill(evInfo.GetQvector(kFHCal_R).GetPhi(), cent);
    hPsiEpVsCent[kFHCal_L]->Fill(evInfo.GetQvector(kFHCal_L).GetPhi(), cent);
    hPsiEpVsCent[kFHCal]->Fill(evInfo.GetQvector(kFHCal).GetPhi(), cent);

    hQxRLVsCent[NetaGaps]->Fill(evInfo.GetQvector(kFHCal_R).X() - evInfo.GetQvector(kFHCal_L).X(), cent);
    hQyRLVsCent[NetaGaps]->Fill(evInfo.GetQvector(kFHCal_R).Y() - evInfo.GetQvector(kFHCal_L).Y(), cent);

    hPsiEpRLVsCent[NetaGaps]->Fill(RangeAngle(evInfo.GetQvector(kFHCal_R).GetPhi() - evInfo.GetQvector(kFHCal_L).GetPhi()), cent);
    hPsiRpRVsCent[NetaGaps]->Fill(RangeAngle(evInfo.GetQvector(kFHCal_R).GetPhi() - mcEvent->GetPhiRP()), cent);
    hPsiRpLVsCent[NetaGaps]->Fill(RangeAngle(evInfo.GetQvector(kFHCal_L).GetPhi() - mcEvent->GetPhiRP()), cent);
  }

  // Init output file
  TFile *fo = new TFile(oFileName.Data(), "recreate");
  fo->cd();
  for (int i = 0; i < NsubEvents; i++)
  {
    hQxVsCent[i]->Write();
    hQyVsCent[i]->Write();
    hPsiEpVsCent[i]->Write();
  }
  for (int i = 0; i < NetaGaps + 1; i++)
  {
    hQxRLVsCent[i]->Write();
    hQyRLVsCent[i]->Write();
    hPsiEpRLVsCent[i]->Write();
    hPsiRpRVsCent[i]->Write();
    hPsiRpLVsCent[i]->Write();
  }
  fo->Close();

  timer.Stop();
  timer.Print();
  return 0;
}